# This file may be edited by the user.
# Please read AUTH.md first before changing anything.

# Method 1. Set this variable here or set it as an environment variable
GCP_PROJECT = "Name of your BigQuery project"


# Method 2. Set this variable here or set it as an environment variable
GCP_CREDENTIALS_PATH = "Path to your service account credentials json file"

